package edu.gmu.cs475;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.Timeout;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

public class Part1FunctionalTests {
	@Rule
	public Timeout globalTimeout = new Timeout(1000);
	private AbstractKeyValueStore kvStore;

	static void assertListContainsOnly(List<String> l, String... objs) {
		boolean failed = false;
		if (objs.length != l.size())
			failed = true;
		else {
			List<String> dup = new ArrayList<>(l);
			for (String s : objs)
				if (!dup.remove(s)) {
					failed = true;
					break;
				}
		}
		if (failed) Assert.fail("Expected list to have: " + Arrays.toString(objs) + ", but found " + l);

	}

	@Before
	public void setup() {
		kvStore = new KeyValueStore();
	}

	@Test
	public void testAddAndGetString() {
		String key = "testAddAndGetStringKey";
		String val = "testAddAndGetStringKeyvalue1";
		kvStore.set(key, val);
		Assert.assertEquals(val, kvStore.get(key));
		Assert.assertEquals(AbstractKeyValueStore.ValueType.STRING, kvStore.getType(key));
	}

	@Test
	public void testAddAndDeleteAndGetString() {
		String key = "testAddAndDeleteAndGetStringKey";
		String val = "testAddAndDeleteAndGetStringKeyvalue1";
		kvStore.set(key, val);
		Assert.assertEquals(val, kvStore.get(key));
		Assert.assertEquals(AbstractKeyValueStore.ValueType.STRING, kvStore.getType(key));
		Assert.assertTrue(kvStore.remove(key));
		Assert.assertEquals(null, kvStore.get(key));
		Assert.assertEquals(null, kvStore.getType(key));
	}

	@Test
	public void testAddAndGetList() {
		String key = "testAddAndGetList";
		ArrayList<String> lst = new ArrayList<>();
		lst.add("foo");
		kvStore.set(key, lst);
		Assert.assertEquals(AbstractKeyValueStore.ValueType.LIST, kvStore.getType(key));
		Assert.assertEquals(lst, kvStore.get(key));
	}

	@Test
	public void testAddAndGetListOverString() {
		String key = "testAddAndGetListOverString";
		ArrayList<String> lst = new ArrayList<>();
		lst.add("foo");
		kvStore.set(key, "test");
		kvStore.set(key, lst);
		Assert.assertEquals(AbstractKeyValueStore.ValueType.LIST, kvStore.getType(key));
		Assert.assertEquals(lst, kvStore.get(key));
		Assert.assertTrue(kvStore.remove(key));
		Assert.assertEquals(null, kvStore.get(key));
		Assert.assertEquals(null, kvStore.getType(key));
	}

	@Test
	public void testAddAndGetStringOverList() {
		String key = "testAddAndGetStringOverList";
		String val = key + "$value";
		ArrayList<String> lst = new ArrayList<>();
		lst.add("foo");
		kvStore.set(key, lst);
		kvStore.set(key, val);
		Assert.assertEquals(val, kvStore.get(key));
		Assert.assertEquals(AbstractKeyValueStore.ValueType.STRING, kvStore.getType(key));
		Assert.assertTrue(kvStore.remove(key));
		Assert.assertEquals(null, kvStore.get(key));
		Assert.assertEquals(null, kvStore.getType(key));
	}

	@Test
	public void testNoNullKeysOrValues() {

		try {
			kvStore.get(null);
			Assert.fail("Should not be able to call get(null) without a NullPointerException");
		} catch (NullPointerException ex) {
		}

		try {
			kvStore.getType(null);
			Assert.fail("Should not be able to call getType(null) without a NullPointerException");
		} catch (NullPointerException ex) {
		}

		try {
			kvStore.set(null, "abcd");
			Assert.fail("Should not be able to call set(null,...) without a NullPointerException");
		} catch (NullPointerException ex) {
		}

		try {
			kvStore.set("abcd", null);
			Assert.fail("Should not be able to call set(...,null) without a NullPointerException");
		} catch (NullPointerException ex) {
		}

		try {
			kvStore.remove(null);
			Assert.fail("Should not be able to call remove(null) without a NullPointerException");
		} catch (NullPointerException ex) {
		}

		try {
			kvStore.addToList(null, "abcd");
			Assert.fail("Should not be able to call addToList(null,...) without a NullPointerException");
		} catch (NullPointerException ex) {
		}


		try {
			kvStore.deleteFromList(null, "abcd");
			Assert.fail("Should not be able to call deleteFromList(null,...) without a NullPointerException");
		} catch (NullPointerException ex) {
		}

	}

	@Test
	public void testCantSetInvalidValue() {

		try {
			kvStore.set("Def", new HashMap<>());
			Assert.fail("Should not be able to call set(...,new HashMap()) without a IllegalArgumentException");
		} catch (IllegalArgumentException ex) {
		}

	}

	@Test
	public void testArrayAddRemoveExceptions() {
		try {
			kvStore.addToList("bah", "abcd");
			Assert.fail("Should not be able to call addToList(invalidKey,...) without a IllegalArgumentException");
		} catch (IllegalArgumentException ex) {
		}

		kvStore.set("bah", "def");
		try {
			kvStore.addToList("bah", "abcd");
			Assert.fail("Should not be able to call addToList(validKey,...) without a IllegalArgumentException");
		} catch (IllegalArgumentException ex) {
		}


		kvStore.remove("bah");
		try {
			kvStore.deleteFromList("bah", "abcd");
			Assert.fail("Should not be able to call deleteFromList(invalidKey,...) without a IllegalArgumentException");
		} catch (IllegalArgumentException ex) {
		}

		kvStore.set("bah", "defA");
		try {
			kvStore.deleteFromList("bah", "abcd");
			Assert.fail("Should not be able to call deleteFromList(validKey,...) without a IllegalArgumentException");
		} catch (IllegalArgumentException ex) {
		}
	}

	@Test
	public void testArrayAddRemove() {
		ArrayList<String> l = new ArrayList<>();
		l.add("v1");
		kvStore.set("key", l);
		kvStore.addToList("key", "v2");
		kvStore.addToList("key", "v2");
		assertListContainsOnly((List<String>) kvStore.get("key"), "v1", "v2", "v2");
		kvStore.deleteFromList("key", "v2");
		assertListContainsOnly((List<String>) kvStore.get("key"), "v1", "v2");

	}

}
